;; Copyright (c) 2021 Sean Cole <scole@NetBSD.org>
;; All rights reserved.
;;
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions
;; are met:
;;
;; 1. Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;; 2. Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the distribution.
;;
;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
;; TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
;; PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
;; BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
;; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
;; SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
;; INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
;; CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
;; ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
;; POSSIBILITY OF SUCH DAMAGE.

;; utilities for working with thai text in emacs buffers.  includes
;; functions to split/unsplit thai strings, thai time conversion,
;; download and play mp3 audio for thai words from thai-language.com,
;; clickthai-online.com, thai2english.com, and lingopolo.org, and currency
;; conversions from x-rates.com.  Also, it can look up thai words if
;; vocabulary files are created.  The 'customize interface is available
;; for some settings.
;;
;; HOW TO USE
;; - in ~/.emacs/init.el, add library path and a require, e.g.:
;;     (add-to-list 'load-path "ST_SHARE_DIR")
;;     (require 'pthai)
;;   or load module directly:
;;     (load-file "ST_SHARE_DIR/pthai.el")
;;
;; - also in ~/.emacs/init.el, possibly bind keys to some commonly
;;   used functions:
;;    (global-set-key [f7] 'pthai-lookup-at-point)
;;    (global-set-key [f8] 'pthai-say-word-at-point)
;;    (global-set-key [f9] 'pthai-say-string-at-point)
;;    (global-set-key [f10] 'pthai-say-line)
;;    (global-set-key [s-right] 'forward-thaiword)
;;    (global-set-key [s-left] 'backward-thaiword)
;;
;; - 'customize should be used to add paths to dictionaries
;; - M-x apropos pthai for available functions
;;
;; TODO
;; - play audio for individual vowels
;; - add routine for practicing vocabulary words or dictionary file
;; - when splitting, handle unknown/misspelled words better (สรี/โกว/หูย)
;; - when parsing thai words, need separate some thai punctuation like "พีฺฺฺ่"
;; - look up word definitions on the fly
;; - interface with pkgsrc/textproc/translate-shell, licensing (?)
;; - handle ๆ/" better, repeating last word and word breaking (?)
;; - with 'pthai-rwb use non-brute-force/dynamic-programming algorithm (?)
;; - create an info file for this module
;; - byte compile module
;; - when doing pthai-lookup, add "classifier" to definition if word classifier (?)
;; - fix pthai-lookup to return classifier info if word only has class info
;; - loading module 2x messes up word counts/defvar problem with 'pthai-dictionary-list?
;; - add 'pthai-mode with menu/keybindings
;; - tab-completion for some functions pthai-say, pthai-lookup (?)
;; - add function to look up file & line numbers where a word is defined
;; - spellchecking for word, buffer, region (string-distance)
;;
;; known issues:
;; - doesn't always handle "long" strings well
;; - when linting or compiling file, need to "(require 'seq)" first(?)
;;
(require 'cus-edit)   ;; custom* customize*
(require 'ido)        ;; ido-completing-read*
(require 'mule-util)  ;; nested alist functions
(require 'pulse)      ;; pulse* functions
(require 'seq)        ;; seq-* functions   
(require 'sort)       ;; sort-lines
(require 'subr-x)     ;; string-trim* functions
(require 'thai-word)  ;; thai-word-table
(require 'thingatpt)  ;; thing-at-point*
(require 'url)        ;; url-* functions

;; xxx "special" vars to set before loading module, not the emacs way?
(unless (boundp 'pthai-bootstrap)
  (defvar pthai-bootstrap nil "nil unless building for pkgsrc"))

(unless (boundp 'pthai-verbose-wordloads)
  (defvar pthai-verbose-wordloads t
    "if non-nil, display word counts when loading dictionaries"))

(defvar pthai-default-directory (concat user-emacs-directory "pthai/")
  "default pthai directory (ensure ends with directory separator)")

(defvar pthai-dictionary (make-hash-table :test 'equal)
   "hash of thai dictionary words.  each word has a dictionary entry, \n\
the form of the dictionary entry is:\n\
thaiword => '( definition english_classifier thai_classisifier)\n\ 
which are all lists of strings.  empty definitions should be defined as nil:\n\
\"ไก่\" => ( '(\"chicken\") nil '(\"ตัว\") ) \n\n\
if a thai word is known but no definition has been created \n\
 '( nil nil nil ) \nwill be its entry ")

(defvar pthai-misc-punctuation-regexp
  (regexp-opt
   (list "~" "`" "!" "@" "#" "\$" "%" "^" "&" "*" "(" ")"
	 "-" "_" "=" "+" "\\" "|" "{" "}" "[" "]"
	 ";" ":" "'" "\"" "<" ">" "." "," "/" "?"
	 "ๆ" "ฯาฯ" "ฯ" "฿" "๏" "๚" "๛"))
  "regexp of misc punctuation used for word splitting")

(defvar pthai-rwb-tmp nil "temporary variable for pthai-rwb") 

(defgroup pthai nil
  "Pthai dictionary, word-splitting and audio-playing for Thai words"
  :group 'applications)

(defcustom pthai-split-mode "biggest"
  "Type of word splitting"
  :group 'pthai
  :set (lambda (sym val) (set-default sym val))
  :type '(radio (const :tag "biggest words possible/fewest words" :value "biggest")
		(const :tag "smallest words possible/most words" :value "smallest")
		(const :tag "interactively display choices" :value "interactive")))

(defcustom pthai-splitter-thai-break-words t
  "use 'thai-break-words from 'thai-word module for word splitting"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-splitter-st-icu t
  "use external program st-icu for word splitting"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-splitter-st-swath t
  "use external program st-swath for word splitting"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-splitter-st-wordbreak t
  "use external program st-wordbreak for word splitting"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-splitter-pthai-rwb t
  "use 'pthai-rwb for word splitting"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-splitter-max-word-length 60
  "maximum string length of thai string used by some programs
when splitting words.  some programs using recursion may not
handle long strings well.  if a string is longer than this
setting, programs using recursion will not try to split it"
  :group 'pthai
  :type 'integer)

(defcustom pthai-fetch-clickthai t
  "fetch mp3 audio from clickthai-online.com"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-fetch-lingopolo t
  "fetch mp3 audio from lingopolo.com"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-fetch-thai2english t
  "fetch mp3 audio from thai2english.com"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-fetch-thailanguage t
  "fetch mp3 audio from thai-language.com"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-audio-download t
  "attempt to download audio files for thai words. if disabled,
only already downloaded files will be played"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-audio-display-definition t
  "when playing thai audio also display definition of word if
true"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-audio-play t
  "play audio files for thai words. if nil, only the definitions
will be displayed"
  :group 'pthai
  :type 'boolean)

(defcustom pthai-audio-strip-silence t
  "when playing audio words strip the silences from them.  this
speeds up playing and makes playing multiple words sound more
natural"
  :group 'pthai
  :type 'boolean)

(defun pthai-directory-create(dir)
  "prompt user to create a directory if it doesn't exist"
  (unless (file-directory-p dir)
    (if (yes-or-no-p (concat dir " does not exist, create?"))
	(or (make-directory dir)
	    (error "failed to create dir %s" dir)))))

(defcustom pthai-audio-directory
  (concat pthai-default-directory "audio")
  "directory where mp3 audio files are download and saved to, as
well as read from"
  :group 'pthai
  :set (lambda (sym val)
	 (set-default sym val)
	 (unless load-in-progress
	   (pthai-directory-create val)))
  :type 'directory)

(defun pthai-randomize(lst)
  "take a list and return it in random order"
  (let (rlst n)
    (while lst
      (setq n (random (length lst)))
      (setq rlst (cons (nth n lst) rlst))
      ;; drop nth element
      (setq lst (append (seq-take lst n) (seq-drop lst (1+ n)))))
    rlst))

;; utilities to manipulate the 'thai-word-table from the 'thai-word
;; library.  these functions should be merged into the 'thai-word
;; module itself, but that module looks to have been written when
;; 'thai-tis620 encoding was commonly used instead of utf-8 for thai
;;
;; 'thai-word-table uses a nested-alist from mule-util which looks
;; like this:
;;  '(3585 1                       ;; ก      word   ก
;;    (3591 1                      ;;  ง     word   กง
;;          (3585 t                ;;    ก               
;;                (3634 t          ;;     า              
;;                      (3619 1))));;      ร word   กงการ
;;    (3585 1                      ;;  ก     word   กก 
;;          (3621 1))))            ;;   ล    word   กกล
;;
(defun pthai-twt-extract(nlist thaistr)
  "helper function to reconstruct thai words from a nested alist,
uses recursion. returns a list of thaiwords from place in 'thai-word-table"
  (let (ucode complete)
    (cond
     ;; finished
     ((not nlist)
      nil)
     ;; (3591 1 ...
     ((integerp (car nlist))
      ;; xxx care about coding-system vars here?
      (setq ucode (char-to-string (car nlist)))
      (setq complete (cadr nlist))
      (setq thaistr (concat thaistr ucode))
      (cond
       ;; t => no word at this depth 
       ((equal complete t)
	(pthai-twt-extract (cddr nlist) thaistr))
       ;; 1 => word at this depth
       ((equal complete 1)
	(append (list thaistr)
		(pthai-twt-extract (cddr nlist) thaistr)))
       (t
	(error "invalid parsing for complete var"))))
     ;; not finished
     (t
      (append (pthai-twt-extract (car nlist) thaistr)
	      (pthai-twt-extract (cdr nlist) thaistr))))))

(defun pthai-twt-p(thaiword)
 "return t if thaiword is in 'thai-word-table, nil otherwise"
  (let ((result
	 (lookup-nested-alist
	  thaiword thai-word-table (length thaiword) 0 t)))
    (and (listp result)
	 (integerp (car result))
	 (= 1 (car result)))))

(defun pthai-twt-count(&optional alist)
  "Count number of words in a nested alist. if no arg given,
count 'thai-word-table words.  message user when called
interactively"
  (interactive)
  (let (elem thaiwords (count 0))
    ;; default list or not
    (setq alist (or alist thai-word-table))
    (unless (nested-alist-p alist)
      (error "Invalid argument %s" alist))
    ;; remove 'thai-words from 'thai-word-table
    (setq alist (cdr alist))
    (while (setq elem (car alist))
      (setq alist (cdr alist))
      (setq thaiwords (pthai-twt-extract elem ""))
      (setq count (+ count (length thaiwords))))
    (if (called-interactively-p 'any)
	(message "%d words" count))
    count))

(defun pthai-twt-read(&optional clear)
  "read 'thai-word-table from 'thai-word into 'pthai-dictionary,
optionally clearing first. return total words in dictionary after
reading"
  (interactive)
  (let ((t_twt thai-word-table))
    (if clear
	(clrhash pthai-dictionary))
    ;; process per letter to help reduce memory usage
    (while (setq t_twt (cdr t_twt))
      (dolist (tmp (pthai-twt-extract (car t_twt) ""))
	(if (string-match-p "^\\ct+$" tmp)
     	    (pthai-dictionary-add-word tmp))))
    (pthai-twt-count)))

(defun pthai-twt-save(filename &optional alist)
  "save thai words extracted from a nested-alist table to
filename in utf8 format, one word per line.  default is to save
the 'thai-word-table if no alist argument given.  Returns number
of dictionary words."
  (interactive "FName of file to save to: \nP")
  (let (thaiwords
	elem
        (coding-system-for-read 'utf-8)
	(coding-system-for-write 'utf-8)
	(buffer-file-coding-system 'utf-8))
    ;; default list or not
    (setq alist (or alist
		    thai-word-table))
    (unless (nested-alist-p alist)
      (error "Invalid argument %s" alist))
    ;; remove 'thai-words from 'thai-word-table alist
    (setq alist (cdr alist))
    ;; process per-letter list one at a time to save memory
    (with-temp-buffer
      (while (setq elem (car alist))
	(setq alist (cdr alist))
	(setq thaiwords (pthai-twt-extract elem ""))
    	(dolist (elem thaiwords)
	  (insert elem "\n")))
      (sort-lines nil (point-min) (point-max))
      (write-region nil nil filename)
      (count-lines (point-min) (point-max)))))

(defun pthai-twt-update(file &optional append)
  "Update Thai word table by replacing the current word list with
FILE, which is utf-8.  If called with optional append argument,
FILE is appended instead to the current word list.  Does the same
as 'thai-update-word-table, except that function expects
'thai-tis620 encoding which appears to be the default format for
at least emacs version <= 28.0"
  (interactive "FThai word table file: \nP")
  (let* ((coding-system-for-read 'utf-8)
	 (coding-system-for-write 'utf-8)
	 (buffer-file-coding-system 'utf-8)
	 (temp_file (make-temp-file "pthai-twt-update-620_")))
    (unwind-protect
	(with-temp-buffer
	  (insert-file-contents file)
	  (setq coding-system-for-write 'thai-tis620)
	  (write-file temp_file))
      (thai-update-word-table temp_file append)
      (delete-file temp_file)
      (if pthai-verbose-wordloads
	  (message "%d words in 'thai-word-table" (pthai-twt-count)))
      thai-word-table)))

(defun pthai-thai-break-words(text)
  "split thai string using 'thai-break-words from 'thai-word
library. returns list of lines with word-strings separated by
spaces"
  (with-temp-buffer
    (insert text)
    (goto-char (point-min))
    (thai-break-words " ")
    (split-string (buffer-string) "\n" t)))

(defun pthai-rwb(thaistr)
  "return list of space separated string of possible word breaks
for a thai string, uses recursion.  can only handle thai words"
  (cond ((<= (length thaistr) 0)
	 nil)
	((string-match-p "\\Ct" thaistr)
	 (error "cannot split non-thai strings: %s" thaistr))
	(t (let ((retval))
	     (setq thaistr (replace-regexp-in-string "[[:space:]]+" "" thaistr))
	     (setq pthai-rwb-tmp nil)
	     (pthai-rwb-helper thaistr (length thaistr) "")
	     (setq retval pthai-rwb-tmp)
	     (setq pthai-rwb-tmp nil)
	     (or retval
		 (list thaistr))))))

(defun pthai-rwb-helper(str n result)
  "recursive helper function for pthai-rwb"
  (let ((i 1)
	prefix
	ss)
    (catch 'done
      (while (<= i n)
	(setq prefix (substring-no-properties str 0 i))
	(when (pthai-dictionary-p prefix)
	  (when (= i n)
	    (setq result (concat result prefix))
	    (setq pthai-rwb-tmp (cons result pthai-rwb-tmp))
	    (throw 'done nil))
	  (setq ss (substring-no-properties str i nil))
	  (pthai-rwb-helper ss (length ss) (concat result prefix " ")))
	(setq i (1+ i))))))

(defun pthai-cycle-split-mode()
  "cycle to next word splitting mode. return string of current
mode. message user when called interactively"
  (interactive)
  (setq pthai-split-mode (downcase pthai-split-mode))
  (cond ((string= pthai-split-mode "biggest")
	 (setq pthai-split-mode "smallest"))
	((string= pthai-split-mode "smallest")
	 (setq pthai-split-mode "interactive"))
	((string= pthai-split-mode "interactive")
	 (setq pthai-split-mode "biggest"))
	(t
	 (error "unknown split mode %s" pthai-split-mode)))
  (if (called-interactively-p 'any)
      (message pthai-split-mode))
  pthai-split-mode)

(defun pthai-split-command(command text)
  "run external command to split thai text.  command should be
whitespace separated commandline arguments. returns list of
strings where each string is a group of words separated by
spaces"
  (let (exe rlist)
    (setq exe (string-trim (car (split-string command))))
    (unless (executable-find exe)
      (error "%s not found" exe))
    (setq rlist
	  (split-string
	   (shell-command-to-string
	    (concat command " \"" text "\"")) "\n" t))))

(defun pthai-split-all(text)
  "function to split up thai text string using various
programs. only parses thai strings with no whitespace.  returns
all found word segmentation combinations as a list of strings,
where each string is a group of words separated by spaces."
  (let (slist
	(ok_length (<= (length text) pthai-splitter-max-word-length)))
    (when (> (length text) 0)
      (unless (string-match-p "^\\ct+$" text)
	(error "can only parse thai strings"))
      (setq slist
	    (append
	     (if pthai-splitter-thai-break-words
		 (pthai-thai-break-words text))
	     (if pthai-splitter-st-icu
		 (pthai-split-command "st-icu" text))
	     (if pthai-splitter-st-swath
		 (pthai-split-command "st-swath" text))
	     ;; don't do recursive algorithms on "long" strings
	     (if (and ok_length pthai-splitter-st-wordbreak)
		 (pthai-split-command "st-wordbreak -a " text))
	     (if (and ok_length pthai-splitter-pthai-rwb)
		 (pthai-rwb text)))))
    (setq slist (remove nil (seq-uniq slist)))))

(defun pthai-split-best(text &optional smode)
  "split up thai string into words.  words are broken up
either by biggest, smallest or interactively depending on
'pthai-split-mode setting and words available in
'pthai-dictionary.  split mode can be specified with optional arg
smode using same values as 'pthai-split-mode setting.  returns a
string containing words separated by spaces"
  (let* ((smode (or smode pthai-split-mode))
	 (slist
	  (sort (pthai-split-all text)
		(lambda (a b) (< (length (split-string a))
    				 (length (split-string b))))))
	 (slen (length slist))
	 (shead (car slist)))
    (cond
     ;; error
     ((< slen 1)
      text)
     ;; only 1 result
     ((= slen 1)
      shead)
     ;; when ordering by biggest words, and 1st word on list is a
     ;; single word, not in dictionary, and not thai number, take the
     ;; 2nd choice of words
     ((and (string= smode "biggest")
	   (not (string-match-p "^[๐๑๒๓๔๕๖๗๘๙]+" shead))
	   (= (length (split-string shead)) 1)
	   (not (pthai-dictionary-p shead)))
      (nth 1 slist))
     ;; first word on list
     ((string= smode "biggest")
      shead)
     ;; last word on list
     ((string= smode "smallest")
      (last slist))
     ;; interactively
     ((string= smode "interactive")
      (ido-completing-read "split words: " slist))
     (t
      shead))))

(defun pthai-bounds-of-thaistring-at-point(&optional cpoint)
  "return beginning and end points of thai string or nil. no word
splitting is done"
  (save-excursion
    (if cpoint 
	(goto-char cpoint))
    (if (looking-at-p "\\Ct")
	nil
      (let ((o_pt (point))
	    (p1 (line-beginning-position))
	    (p2 (line-end-position)))
	(if (re-search-backward "\\(\\Ct\\)" (line-beginning-position) t)
	    (setq p1 (match-end 1)))
	(goto-char o_pt)
	(if (re-search-forward "\\(\\ct+\\)" (line-end-position) t)
	    (setq p2 (match-end 1)))
	(cons p1 p2)))))

(defun pthai-bounds-of-thaiwords-at-point(&optional cpoint)
  "return a list of word bounds for a contiguous thai string at
point. does pthai-word splitting to calculate bounds.  returns
nil if point not at thai string"
  (save-excursion
    (let* ((tmp (pthai-bounds-of-thaistring-at-point (or cpoint (point))))
	   (p1 (car tmp))
	   (p2 (cdr tmp))
	   s_len s_off)
      (when (and p1 p2)
	(setq s_len
	      (mapcar 'length
		      (split-string
		       (pthai-split-best (buffer-substring-no-properties p1 p2)))))
	(setq s_off (list 0))
	(dolist (tmp (butlast s_len))
	  (setq s_off (cons (+ tmp (car s_off)) s_off)))
	(setq s_off (seq-map-indexed
		     (lambda (x n) (cons (+ p1 x)
					 (+ p1 x (nth n s_len))))
		     (nreverse s_off))))
      s_off)))

(defun pthai-bounds-of-thaiword-at-point(&optional cpoint)
  "Return beginning and end boundaries of a thai word at point.  does
word-splitting to determine the actual thai word.  used with
thingatpoint library to allow 
  (thing-at-point 'thaiword)
  (forward-thaiword)
  (backword-thaiword)
  (beginning-of-thing 'thaiword)
  (end-of-thing 'thaiword)
to work (others?).  returns nil if boundaries not found"
  (save-excursion
    (let* ((pt (or cpoint (point)))
	   (bounds (pthai-bounds-of-thaiwords-at-point pt))
	   (bound (and bounds
		       (seq-find
			(lambda (x) (and (>= pt (car x)) (< pt (cdr x))))
			bounds nil))))
      bound)))

;; add to thingatpoint library
(put 'thaiword 'bounds-of-thing-at-point
     'pthai-bounds-of-thaiword-at-point)

(defun pthai-beginning-of-thaiword()
  "move point to beginning of thaiword"
  (beginning-of-thing 'thaiword))

(defun pthai-end-of-thaiword()
  "move point to end of thaiword"
  (end-of-thing 'thaiword))

(defun forward-thaiword(&optional arg)
  "Move point forward 1 or arg thaiwords (backward if arg is
negative). returns t if moved past a thai word, else nil.
skips over non-thai text.  (see also 'thai-forward-word)"
  (interactive "p")
  (let ((arg (or arg 1))
	(success t)
	tmp)
    ;; backward - move to beginning of closest thaiword backward from point
    (while (and (< arg 0) success)
      (cond
       ((bobp)
	(setq success nil))
       ((looking-at-p "\\ct")
	(setq tmp (car (pthai-bounds-of-thaiword-at-point)))
	(when (= (point) tmp)
	  (setq tmp (1- tmp))
	  (setq arg (1- arg)))
	(goto-char tmp))
       (t
	(unless (and (re-search-backward "\\ct" nil t)
		     (goto-char (car (pthai-bounds-of-thaiword-at-point))))
	  (setq success nil))))
      (setq arg (1+ arg)))
    ;; forward - move to end of closest thai word going right
    (while (and (> arg 0) success)
      (cond
       ((eobp)
	(setq success nil))
       ((looking-at-p "\\ct")
	(goto-char (cdr (pthai-bounds-of-thaiword-at-point))))
       (t
	(unless (and (re-search-forward "\\ct" nil t)
		     (goto-char
		      (cdr (pthai-bounds-of-thaiword-at-point (1- (point))))))
	  (setq success nil))))
      (setq arg (1- arg)))
    success))

(defun backward-thaiword(&optional arg)
  "move backward until encountering the beginning of an thai
  word.  with arg, do arg times. (see also 'thai-forward-word)"
  (interactive "p")
  (let ((arg (or arg 1)))
    (forward-thaiword (- 0 arg))))

(defun pthai-split-region(&optional p1 p2)
  "in a region defined by p1 and p2, split thai words into
subwords separated by spaces"
  (interactive)
  (cond
   ((and p1 p2)
    nil)
   ((use-region-p)
    (setq p1 (region-beginning))
    (setq p2 (region-end)))
   (t
    (error "no region selected")))
  (let (bounds (o_pt (point)))
    (save-restriction
      (narrow-to-region p1 p2)
      (goto-char p1)
      ;; divide up thai words with no spaces between
      (while (re-search-forward "\\(\\ct+\\)" nil t)
	(when (and (setq bounds
			 (pthai-bounds-of-thaiwords-at-point (1- (point))))
		   (> (length bounds) 1))
	  (insert
	   (string-join
	    (mapcar (lambda (x) (buffer-substring (car x) (cdr x))) bounds)
	    " "))
	  (delete-region (caar bounds)
			 (cdar (last bounds)))))
      ;; append spaces
      (goto-char p1)
      (while (re-search-forward "\\(\\ct\\)\\(\\Ct\\)" nil t)
	(unless (string-match-p "[[:space:]]" (match-string 2))
	  (replace-match (concat (match-string 1) " " (match-string 2)))))
      ;; prepend spaces
      (goto-char p1)
      (while (re-search-forward "\\(\\Ct\\)\\(\\ct\\)" nil t)
	(unless (string-match-p "[[:space:]]" (match-string 1))
	  (replace-match (concat (match-string 1) " " (match-string 2))))))
    ;; move point to beginning leave at end of modified region
    ;; depending where it was originally
    (if (= o_pt p1)
	(goto-char o_pt))))

(defun pthai-split(text &optional smode)
  "split thai text.  return split string of text.  if optional
smode, override current pthai-split-mode"
  (interactive "stext: ")
  (let (stext)
    (with-temp-buffer
      (insert text)
      (goto-char (point-min))
      (pthai-split-region (point-min) (point-max))
      (setq stext (buffer-substring (point-min) (point-max))))
    (if (called-interactively-p 'any)
	(message "%s" stext))
    stext))

(defun pthai-split-string-at-point()
  "split thai string at point, returns nil if point not at thai
string"
  (interactive)
  (let* ((bounds (pthai-bounds-of-thaistring-at-point))
	 (p1 (car bounds))
	 (p2 (cdr bounds)))
    (if (and p1 p2)
	(pthai-split-region p1 p2)
      nil)))

(defun pthai-split-line()
  "split thai words on current line"
  (interactive)
  (pthai-split-region (line-beginning-position)
		      (line-end-position)))

(defun pthai-split-line-from-point()
  "split thai words from point, or beginning of thai word at point,
to the end of the line"
  (interactive)
  (pthai-split-region
   (or (car (pthai-bounds-of-thaiword-at-point)) (point))
   (line-end-position)))

(defun pthai-split-buffer()
  "split thai words in buffer"
  (interactive)
  (pthai-split-region (point-min) (point-max)))

(defun pthai-unsplit-region(&optional p1 p2)
  "remove space and tabs from thai words in region"
  (interactive)
  (let* ((nonnum_rexp "\\([\u0e00-\u0e4f\u0e5a-\u0e7f]\\)")
	 (space_regexp
	  (concat nonnum_rexp "\\([ \t]+\\)" nonnum_rexp)))
    (cond
     ((and p1 p2)
      nil)
     ((use-region-p)
      (setq p1 (region-beginning))
      (setq p2 (region-end)))
     (t
      (error "no region selected")))
    (save-excursion
      (save-restriction
	(narrow-to-region p1 p2)
	(goto-char (point-min))
	;; remove spaces between thai words excluding numbers
	(while (re-search-forward space_regexp nil t)
	  (replace-match (concat (match-string 1) (match-string 3))))))))

(defun pthai-unsplit-line()
  "remove spaces from thai words on line starting at point"
  (interactive)
  (pthai-unsplit-region (point) (line-end-position)))

(defun pthai-unsplit-buffer()
  "remove spaces from thai words in buffer"
  (interactive)
  (pthai-unsplit-region (point-min) (point-max)))

(defun pthai-url-clickthai(thaiword)
  "return mp3 url for a thai word from clickthai-online.com or
nil if unavailable"
  (let ((url "https://www.clickthai-online.com/wbtde/dictionary.php")
	(url-request-method "POST")
	(url-request-extra-headers
	 `(("Content-Type" . "application/x-www-form-urlencoded; charset=utf-8")))
	(url-request-data
	 (concat "action=search&search=" (url-encode-url thaiword)))
	(play_url "http://www.clickthai-online.com/cgi-bin/playmp3.pl?")
	sound_url
	tmp
	tmp_url)
    (with-current-buffer (url-retrieve-synchronously url)
      (goto-char (point-min))
      ;; <a href="javascript:play('../cgi-bin/playmp3.pl?322476');">วัน</a>
      (while (and (not sound_url)
		  (re-search-forward "playmp3\\.pl\\?\\([0-9]+\\)');\">\\(.*\\)</a>" nil t))
	(and (setq tmp (decode-coding-string (match-string 2) 'utf-8))
	     (setq tmp_url (concat play_url (match-string 1)))
	     (string= tmp thaiword)
	     (setq sound_url tmp_url))))
    sound_url))

(defun pthai-url-thai-language(thaiword)
  "return mp3 url for a thai word from thai-language.com or nil
if unavailable"
  (let* (title_word
	 ;; <title>thai-language.com - แมว</title>
	 (title_regexp "<title>thai-language.com - \\(.*\\)</title>")
	 id
	 ;; <link rel="canonical" href="http://www.thai-language.com/id/131188" />
	 (id_regexp "<link rel=\"canonical\" href=\"http://www.thai-language.com/id/\\([0-9]+\\)\"")
	 click_regexp
	 sound_url
	 (tmmode 0) ;;  exact=0 contains=1 starts_with=2 ends_with=3
	 (mp3_url "http://thai-language.com/mp3/")
	 (url "http://thai-language.com/dict")
	 (url-request-method "POST")
	 (url-request-extra-headers
	  `(("Content-Type" . "application/x-www-form-urlencoded; charset=utf-8")))
	 (url-request-data
	  (concat "action=/default.aspx&tmode="
		  (number-to-string tmmode) "&search=" (url-encode-url thaiword))))
    (with-current-buffer (url-retrieve-synchronously url)
      (and (goto-char (point-min))
	   (re-search-forward title_regexp nil t 1)
	   (setq title_word (decode-coding-string (match-string 1) 'utf-8)))
      (and (goto-char (point-min))
	   (re-search-forward id_regexp nil t 1)
	   (setq id (match-string 1)))
      (and id
	   (goto-char (point-min))
	   ;; <a onClick=PlayAudioFile('/mp3/E131188.mp3') style='cursor:pointer;'>
	   ;;   <img src='/img/speaker.gif' /></a>
	   (setq click_regexp
		 (concat "a onClick=PlayAudioFile('/mp3/\\([A-Z]" id "\\.mp3\\)') "))
	   (re-search-forward click_regexp nil t 1)
	   (setq sound_url (concat mp3_url (match-string 1)))))
    ;; title_word needs to match thaiword because query may return
    ;; other pages, e.g., "บักหำ" returns page for "หำ"
    (and title_word
	 (string= title_word thaiword)
	 id
	 sound_url)))

(defun pthai-url-thai2english(thaiword)
  "return mp3 url for a thai word from thai2english.com or nil if
unavailable"
  (let* ((t2e_url "https://www.thai2english.com")
	 (fetch_url (url-encode-url (concat t2e_url "/search?q=" thaiword)))
	 (id_match (concat t2e_url "/dictionary/\\([0-9]+\\)\\.html"))
	 (speaker_url "https://ast.thai2english.com/img/circle-icons/speaker.svg")
	 (sounds_url "https://ast.thai2english.com/sounds/"))
    (with-current-buffer (url-retrieve-synchronously fetch_url)
      (goto-char (point-min))
      (and (search-forward speaker_url nil t)
	   (goto-char (point-min))
	   (re-search-forward id_match nil t 1)
	   (concat sounds_url (match-string 1) ".mp3")))))

(defun pthai-url-lingopolo(thaiword)
  "return mp3 url for a thai word from lingopolo.org or nil if
unavailable"
  (let*
      ((lingo_url  "https://lingopolo.org/thai")
       ;; https://lingopolo.org/thai/search/node?keys=เอ้อ, แมว เอ้อ เสือ คน
       (search_url
	(concat lingo_url "/search/node?keys=" (url-encode-url thaiword)))
       ;; xxx this is gross and probably fragile...
       ;; \n[whitespace]<strong>เอ้อ</strong>[whitespace]...</p>
       (yes_match
	(concat "\n[ \t]+<strong>" thaiword "</strong>[ \t\n\r]+\.\.\.</p>"))
       (no_match "Your search yielded no results.")
       ;; <a href="https://lingopolo.org/thai/word/also-well">also; well...</a>
       (found_regexp
	(concat "<a href=\"\\(" lingo_url "/word/.*\\)\">.*</a>"))
       found_url
       ;; <audio src="/thai/sites/lingopolo.org.thai/files/2013/w/well_0.mp3" controls=.*">
       (audio_regexp
	"<audio src=\"\\(/thai/sites/.*\\mp3\\)\" controls=")
       ;; https://lingopolo.org/thai/sites/lingopolo.org.thai/files/2013/w/well_0.mp3
       mp3_url)
    
    (with-current-buffer (url-retrieve-synchronously search_url)
      (set-buffer-multibyte t)
      (and (goto-char (point-min))
	   (not (re-search-forward no_match nil t))
	   (goto-char (point-min))
	   (re-search-forward yes_match nil t)
	   (re-search-backward found_regexp nil t)
	   (setq found_url (match-string 1))))

    (when found_url
      (with-current-buffer (url-retrieve-synchronously found_url)
	(set-buffer-multibyte t)
	(and (goto-char (point-min))
	     (re-search-forward audio_regexp nil t)
	     (setq mp3_url
		   (replace-regexp-in-string "^/thai" lingo_url (match-string 1))))))
    
    mp3_url))

(defun pthai-soundfile(thaiword)
  "generate default sound path to store mp3 audio file for a
thaiword. file may or may not already exist"
  (unless (string-match-p "^\\ct+$" thaiword)
    (error "only thai words supported (%s)" thaiword))
  (unless (and pthai-audio-directory
	       (file-directory-p pthai-audio-directory))
    (error "pthai-audio-directory \"%s\" not found"  pthai-audio-directory))
  (let* ((subdir
	  (concat pthai-audio-directory "/"
		  (number-to-string (string-to-char thaiword))))
	 (file
	  (expand-file-name
	   (concat subdir "/" thaiword ".mp3"))))
    (unless (file-directory-p subdir)
      (make-directory subdir))
    file))

(defun pthai-soundfile-downloaded-p(thaiword)
  "return non-nil if a soundfile download attempted already for a
thai word"
  (file-exists-p (pthai-soundfile thaiword)))

(defun pthai-soundfile-playable-p(thaiword)
  "return non-nil if a soundfile downloaded successfully for a
thai word"
  (and (pthai-soundfile-downloaded-p thaiword)
       (> (file-attribute-size (file-attributes (pthai-soundfile thaiword))) 0)
       t))

(defun pthai-soundfiles-sort(tlist)
  "sort list of list of thai strings in order of most available
audio files as a percentage, or containing the biggest words"
  (sort tlist
	(lambda (a b)
	  (let*
	      ((a_total (length a))
	       (b_total (length b))
	       (a_found (length (seq-filter (lambda (x) x) (mapcar 'pthai-soundfile-playable-p a))))
	       (b_found (length (seq-filter (lambda (x) x) (mapcar 'pthai-soundfile-playable-p b))))
	       (a_percent (round (* 100.0 (/ (float a_found) (float a_total)))))
	       (b_percent (round (* 100.0 (/ (float b_found) (float b_total))))))
	    (cond
	     ;; use biggest length of first word if all rest same
	     ((and (= a_total b_total)
		   (= a_percent b_percent))
	      (> (length (car a)) (length (car b))))
	     ;; use biggest words if percentage same
	     ((= a_percent b_percent)
	      (< a_total b_total))
	     ;; else highest percent of words available
	     (t
	      (> a_percent b_percent)))))))

(defun pthai-mp3-play(thaiword)
  "use sox/play to play audio file for a thai word. attempts to
remove silences from audio file unless 'pthai-audio-strip-silence
is nil.  assumes soundfile already downloaded and in default
sound directory. return t if playing successful, nil or error
otherwise."
  (let ((tpath (pthai-soundfile thaiword))
	(sflags "silence 1 0.1 1% reverse silence 1 0.1 1% reverse")
	tmp)
    (cond
     ;; audio disabled
     ((not pthai-audio-play)
      nil)
     ;; check for executables
     ((or (not (executable-find "sox"))
	  (not (executable-find "play")))
      (error "sox/play not found"))
     ;; check for audiofile
     ((not (pthai-soundfile-playable-p thaiword))
      (error "file not playable or existing: %s" tpath))
     ;; play audio
     (t
      (unless pthai-audio-strip-silence
	(setq sflags nil))
      (setq tmp
	    (shell-command-to-string
	     (concat "play -q \"" tpath "\" " sflags)))
      ;; xxx warnings about clipped samples, sound still ok?
      (if (and tmp (or (string= tmp "")
		       (string-match-p "WARN .*rate clipped" tmp)))
	  t
	(and (message tmp)
	     nil))))))

(defun pthai-mp3-url(thaiword)
  "return mp3 sound url for thai word from random one of thai-language,
clickthai, lingopolo or thai2english, or nil if not
found. returns nil unless thaiword argument given"
  (let ((fetchers
	 (remove nil
		 (list
		  (if pthai-fetch-clickthai    'pthai-url-clickthai)
		  (if pthai-fetch-lingopolo    'pthai-url-lingopolo)
		  (if pthai-fetch-thai2english 'pthai-url-thai2english)
		  (if pthai-fetch-thailanguage 'pthai-url-thai-language))))
	url)
    (when (and (string-match-p "\\ct+" thaiword) fetchers)
      (setq fetchers (pthai-randomize fetchers))
      (while (and (not url) fetchers)
	(setq url (funcall (car fetchers) thaiword))
	(setq fetchers (cdr fetchers))))
    url))

(defun pthai-mp3-fetch(thaiword &optional force_dl)
  "fetch mp3 soundfile to default audio directory if file missing
or forced to. returns path to downloaded file or nil if unable to
download. if download unsuccesful, write empty file with name of
thai word to default path.  if downloading disabled and non-empty
file already downloaded, return full path"
  (let ((file (pthai-soundfile thaiword))
	url)
    (unless (string-match-p "\\ct+" thaiword)
      (error "thaiword %s contains no thai chars"))
    (cond
     ;; file already downloaded but downloading disabled
     ((and (not pthai-audio-download)
	   (pthai-soundfile-playable-p thaiword))
      file)
     ;; downloading enabled, forced download or file not downloaded,
     ;; try to download and return path if successful
     ((and pthai-audio-download
	   (or force_dl (not (pthai-soundfile-downloaded-p thaiword))))
      (if (setq url (pthai-mp3-url thaiword))
	  (and (url-copy-file url file t) file)
	(and (write-region "" nil (pthai-soundfile thaiword))
	     nil)))
     ;; otherwise, no downloading and return whether file available
     (t
      (if (pthai-soundfile-playable-p thaiword)
	  file nil)))))

(defun pthai-download-and-play(word &optional force_dl)
  "download and play audio for a thai word. returns t if played
complete audio for word, nil otherwise"
  (unless (string-match-p "^\\ct+$" word)
    (error "only thai words supported"))
  ;; most common cases: existing word there and available for
  ;; download/download successful
  (if (and (pthai-mp3-fetch word force_dl) (pthai-mp3-play word))
      t
    ;; split up word download and play "best' word combo
    (let* ((splits (pthai-split-all word))
	   (tmp (pthai-soundfiles-sort (mapcar 'split-string splits)))
	   dl_done play_done)
      ;; try to download at one set that makes up the word
      (when pthai-audio-download
	(dolist (tmp (pthai-soundfiles-sort (mapcar 'split-string splits)))
	  (and (not dl_done)
	       (mapc (lambda(x) (pthai-mp3-fetch x force_dl)) tmp)
	       (seq-every-p 'pthai-soundfile-playable-p tmp)
	       (setq dl_done tmp))))
      ;; find combo with most downloads and play
      (when pthai-audio-play
	(setq play_done (or dl_done
			    (car (pthai-soundfiles-sort
				  (mapcar 'split-string splits)))))
	(setq play_done
	      (mapcar (lambda (x)
			(and (pthai-soundfile-playable-p x)
			     ;; skip single consonants
			     (not (string-match-p "^[\u0e00-\u0e2e]\\{1\\}$" x))
			     (pthai-mp3-play x)))
		      play_done)))
      (if (member nil play_done)
	    nil t))))

(defun pthai-highlight-say-bounds(bounds &optional force_dl no_audio)
  "helper to highlight and say word at boundary, lookup and
message its definition. if option force_dl, force audio download.
if option no_audio, don't play audio"
  (let* ((p1 (car bounds))
	 (p2 (cdr bounds))
	 (word (buffer-substring-no-properties p1 p2)))
    (when (and p1 p2 word)
      (cond
       ;; number
       ((string-match-p "^\\([0-9]+\\|[๐๑๒๓๔๕๖๗๘๙]+\\)$" word)
	(pulse-momentary-highlight-region p1 p2)
	(redisplay)
	(pthai-lookup word t)
	(or no_audio (pthai-say-number word)))
       ;; thai word
       ((string-match-p "^\\ct+$" word)
	(pulse-momentary-highlight-region p1 p2)
	(redisplay)
	(pthai-lookup word t)
	(or no_audio (pthai-download-and-play word force_dl)))
       ;; english word
       ((string-match-p "^\\([a-zA-Z]+\\)$" word)
	(pulse-momentary-highlight-region p1 p2)
	(redisplay)
	(if no_audio
	    (pthai-lookup word t)
	  (dolist (tmp (pthai-lookup word t))
	    (pthai-download-and-play tmp force_dl))))))))

(defun pthai-say-region-helper(p_begin p_end &optional force_dl thai_only)
  "helper function to play audio for a selected region. force
downloads with force_dl arg. play only thai words with thai_only
arg"
  (let* ((old_cursor cursor-type)
	 (m_list (list "[๐๑๒๓๔๕๖๗๘๙]+" "[\u0e00-\u0e4f\u0e5a-\u0e7f]+"
		       "[\?\.!]" "[0-9]+"
		       "[a-zA-Z]+"))
	 (m_rexp
	  (concat "\\("
		  (string-join (if thai_only (butlast m_list) m_list) "\\|")
		  "\\)"))
	 bounds)
    (unwind-protect
	(save-mark-and-excursion
	  (goto-char p_begin)
	  (deactivate-mark)
	  (setq cursor-type nil)
	  (while (re-search-forward m_rexp p_end t)
	    (setq bounds
		  (or (pthai-bounds-of-thaiwords-at-point (match-beginning 1))
		      (list (cons (match-beginning 1) (match-end 1)))))
	    ;; filter words that are completely out of the region, which can
	    ;; happen if thai string overlaps end points
	    (setq bounds
		  (seq-remove (lambda (x) (or (>= p_begin (cdr x))
					      (>= (car x) p_end)))
			      bounds))
	    (dolist (tmp bounds)
	      (pthai-highlight-say-bounds tmp force_dl))))
      (setq cursor-type old_cursor))))

(defun pthai-say(text &optional force_dl thai_only)
  "split up text string, download and play audio for thai
words. return definitions of words as a string.  also say arabic
or thai numbers in thai.  see 'pthai-split for more info
on how words are split"
  (interactive "sword: ")
  (let (words)
    (cond
     ;; meanings of english words with spaces, e.g., "good morning"
     ((and (string-match-p "^[a-zA-Z ]+$" text)
	   (setq words (or (pthai-lookup text)
			   (pthai-lookup-classifier text))))
      (message "%s = %s" text (string-join words ", "))
      (dolist (tmp words)
	(pthai-download-and-play tmp force_dl)))
     (t
      (save-excursion
	(with-temp-buffer
	  (insert text)
	  (pthai-say-region-helper (point-min)
				   (point-max) force_dl thai_only)))))))

(defun pthai-say-region(&optional force_dl)
  "play thai audio for highlighted region"
  (interactive "P")
  (unless (use-region-p)
    (error "no region selected"))
  (pthai-say-region-helper (region-beginning) (region-end) force_dl))

(defun pthai-say-region-thai-only(&optional force_dl)
  "play thai audio for thai words in highlighted region"
  (interactive "P")
  (unless (use-region-p)
    (error "no region selected"))
  (pthai-say-region-helper (region-beginning) (region-end) force_dl t))

(defun pthai-say-buffer(&optional force_dl)
  "play thai audio for buffer"
  (interactive "P")
  (pthai-say-region-helper (point-min) (point-max) force_dl))

(defun pthai-say-buffer-thai-only(&optional force_dl)
  "play thai audio for thai words in buffer"
  (interactive "P")
  (pthai-say-region-helper (point-min) (point-max) force_dl t))

(defun pthai-say-line(&optional force_dl)
  "play thai audio for current line in buffer"
  (interactive "P")
  (pthai-say-region-helper (line-beginning-position)
			   (line-end-position)
			   force_dl))

(defun pthai-say-line-thai-only(&optional force_dl)
  "play thai audio for current line in buffer, thai words only"
  (interactive "P")
  (pthai-say-region-helper (line-beginning-position)
			   (line-end-position)
			   force_dl
			   t))

(defun pthai-say-string-at-point(&optional force_dl)
  "play thai audio for thai string (thai text separated by
whitespace) at point in buffer"
  (interactive "P")
  (let* ((tmp (pthai-bounds-of-thaistring-at-point))
	 (p1 (car tmp))
	 (p2 (cdr tmp)))
    (unless (and p1 p2)
      (error "could not determine bounds"))
    (pthai-say-region-helper p1 p2 force_dl)))

(defun pthai-say-word-at-point(&optional force_dl)
  "play thai audio for thing at point. returns error if
bounds cannot be determined"
  (interactive "P")
  (let ((bounds
	 (cond
	  ;; thai word
	  ((looking-at-p "\\ct")
	   (pthai-bounds-of-thaiword-at-point))
	  ;; eng word
	  ((looking-at-p "[A-Za-z]")
	   (bounds-of-thing-at-point 'word))
	  ;; number
	  ((looking-at-p "[0-9๑๒๓๔๕๖๗๘๙๐]")
	   (bounds-of-thing-at-point 'word))
          ;; other, xxx punc?
	  (t
	   nil))))
    (unless bounds
      (error "could not determine bounds"))
    (pthai-highlight-say-bounds bounds force_dl)))

(defun pthai-say-number(num)
 "say number (thai or arabic) in thai. if arabic number, first
convert to thai. returns thai number (string), or nil if can't
convert.  only integers < abs(10 million) are supported.  use
'pthai-say for interactive usage"
  (let* ((powers '(("1" . "เอ็ด") ("10" . "สิบ") ("20" . "ยี่สิบ")
		  ("100" . "ร้อย") ("1000" . "พัน") ("10000" . "หมื่น")
		  ("100000" . "แสน") ("1000000" . "ล้าน")))
	(fn2w (lambda (x) (cdr (assoc x powers))))
	(negative "ลบ")
	(max_num 9999999)
	is_neg
	anum
	rnum
	tlist
	tmp
	power
	n)
    (unless (stringp num)
      (setq num (number-to-string num)))
    ;; negative number
    (when (string-match-p "^-" num)
      (setq num (replace-regexp-in-string "^-" "" num))
      (setq is_neg t))
    (unless (string-match-p "^[0-9๐๑๒๓๔๕๖๗๘๙]+" num)
      (error (concat "not valid number " num)))
    (if (string-match-p "^[๐๑๒๓๔๕๖๗๘๙]+" num)
	(setq num (pthai-number num)))
    (unless (string-match-p "^[0-9]+$" num)
      (setq num (replace-regexp-in-string "[^0-9]" "" num)))
    (setq anum (string-to-number num))
    (cond ((> anum max_num)
	   (warn (concat "audio only for numbers <= abs(" (number-to-string max_num) ")")))
	  ((< anum 10)
	   (setq tlist (list anum)))
	  (t
	   (setq rnum (reverse (split-string (pthai-number num) "" t)))
	   ;; digit 1
	   (setq tmp (nth 0 rnum))
	   (cond ((string= "๑" tmp)
		  (setq tlist (cons (funcall fn2w "1") tlist)))
		 ((not (string= "๐" tmp))
		  (setq tlist (cons tmp tlist))))
	   ;; digit 2
	   (when (setq tmp (nth 1 rnum))
	     (cond ((string= "๑" tmp)
		    (setq tlist (cons (funcall fn2w "10") tlist)))
		   ((string= "๒" tmp)
		    (setq tlist (cons (funcall fn2w "20") tlist)))
		   ((not (string= "๐" tmp))
		    (setq tlist (cons (funcall fn2w "10") tlist))
		    (setq tlist (cons tmp tlist)))))
	   ;; digits 3-7
	   (setq power "10")
	   (dotimes (n 5)
	     (when (setq tmp (nth (+ n 2) rnum))
	       (setq power (concat power "0"))
	       (when (not (string= tmp "๐"))
		 (setq tlist (cons (funcall fn2w power) tlist))
		 (setq tlist (cons tmp tlist)))))))
    ;; convert all to strings
    (setq tlist
	  (mapcar (lambda (x) (if (numberp x) (number-to-string x) x))
		  tlist))
    ;; convert all arabic numbers to thai
    (setq tlist
	  (mapcar (lambda (x) (if (string-match-p "^[0-9]+" x)
				  (pthai-number x) x))
		  tlist))
    ;; negative
    (if is_neg (setq tlist (cons negative tlist)))
    ;; say the number
    (dolist (tmp tlist)
      (pthai-download-and-play tmp))
    ;; return thai number as a string
    (setq tmp (pthai-number num))
    (if is_neg (setq tmp (concat "-" tmp)))
    (if (called-interactively-p 'any)
	(message tmp))
    tmp))

(defun pthai-number(&optional num)
  "convert a number between thai and arabic, and vice versa.
message user converted value if called interactively"
  (interactive)
  (let ((nlist '(("0" . "๐") ("1" . "๑") ("2" . "๒") ("3" . "๓")
		 ("4" . "๔") ("5" . "๕") ("6" . "๖") ("7" . "๗")
		 ("8" . "๘") ("9" . "๙")))
	numa
	res)
    (unless num
      (setq num (read-string "number: ")))
    (unless (stringp num)
      (setq num (number-to-string num)))
    (setq numa (split-string num "" t))
    (if (string-match-p "[0-9]+" num)
	(setq res
	      (mapconcat (lambda(x) (or (cdr (assoc x nlist)) x)) numa ""))
      (setq res
	    (mapconcat (lambda(x) (or (car (rassoc x nlist)) x)) numa "")))
    (if (called-interactively-p 'any)
	(message res))
    res))

(defun pthai-parse-hour-minute-second(timestr &optional thaitime)
  "parse hour-minute-second strings like 7 1pm 12:32pm 22:07
9:45p.m. 0645 3.27am 15:23:55 or \"now\".  returns list of (SEC
MIN HOUR), with nil for each element not parsed"
  (let ((ampm_rexp "\\([AaPp]\.?[Mm]\.?\\)$")
	am_pm second minute hour tlist)
    ;; "now"
    (if (string-match-p "^[Nn][Oo][Ww]$" timestr)
	(setq timestr
	      (nth 3 (split-string
		      (pthai-time-now (not thaitime))))))
    ;; strip out anything that doesn't look like a time
    (setq timestr
	  (replace-regexp-in-string "[^0-9:\.pPaAmM]" "" timestr))
    ;; determine am/pm and remove from string
    (when (string-match-p ampm_rexp timestr)
      (if (string-match-p "[Aa]" timestr)
	  (setq am_pm "am")
	(setq am_pm "pm"))
      (setq timestr
	    (replace-regexp-in-string ampm_rexp "" timestr)))
    ;; parse hour/minute/seconds digits
    (cond
     ;; 1 or 2 digit, 1 11
     ((string-match "^\\([0-9]\\{1,2\\}\\)$" timestr)
      (setq hour (match-string 1 timestr))
      (setq minute "0")
      (setq second "0"))
     ;; 4 digits, 0645
     ((string-match "^\\([0-9]\\{4\\}\\)$" timestr)
      (setq hour (substring timestr 0 2))
      (setq minute (substring timestr 2 4))
      (setq second "0"))
     ;; 6 digits, 064532
     ((string-match "^\\([0-9]\\{6\\}\\)$" timestr)
      (setq hour (substring timestr 0 2))
      (setq minute (substring timestr 2 4))
      (setq second (substring timestr 4 6)))
     ;; 12:32 22:07 3.27 2:35
     ((string-match "^\\([0-9]\\{1,2\\}\\)[:\.]\\([0-9]\\{1,2\\}\\)$" timestr)
      (setq hour (match-string 1 timestr))
      (setq minute (match-string 2 timestr))
      (setq second "0"))
     ;; 12:43:54 or 12.43.54
     ((string-match
       "^\\([0-9]\\{1,2\\}\\)[:\.]\\([0-9]\\{1,2\\}\\)[:\.]\\([0-9]\\{1,2\\}\\)$" timestr)
      (setq hour (match-string 1 timestr))
      (setq minute (match-string 2 timestr))
      (setq second (match-string 3 timestr))))
    ;; convert to strings
    (if second (setq second (string-to-number second)))
    (if minute (setq minute (string-to-number minute)))
    (if hour (setq hour (string-to-number hour)))
    ;; adjust/sanity checks for am/pm
    (when (and am_pm hour)
      (cond
       ((and am_pm (> hour 12))
	(setq hour nil))
       ((and (string= am_pm "am") (= 12 hour))
	(setq hour 0))
       ((and (string= am_pm "pm") (< hour 12))
	(setq hour (+ 12 hour)))))
    ;; return nil unless all seem valid
    (if (and hour (>= hour 0) (<= hour 24)
	     minute (>= minute 0) (< minute 60)
	     second (>= second 0) (< second 60))
	(list second minute hour)
      (list nil nil nil))))

(defun pthai-time-loc2th(timestr &optional reverse)
"convert an hour:minute:second time string to thai time. use
\"now\" for current time.  if optional reverse, convert thai time
to local time"
  (interactive "stime: ")
  (let* ((c_time (current-time))
	 (time_format (concat "%a %b %e %H:%M:%S %Z %Y"))
	 (thai_offset (* 7 3600)) ;; UTC+7 or ICT or +07 or "Asia/Bangkok"
	 (thai_zone (current-time-zone nil thai_offset))
	 tlist enc_time loc_fts thai_fts o_str)
    ;; keep reading until valid second/minute/hour
    (while (not
	    (and (setq tlist (pthai-parse-hour-minute-second timestr reverse))
		 (nth 0 tlist)
		 (nth 1 tlist)
		 (nth 2 tlist)))
      (setq timestr (read-string "time: ")))
    (cond
     ;; thai time
     (reverse
      (setq tlist (append tlist (nthcdr 3 (decode-time c_time thai_zone))))
      (setq enc_time (apply 'encode-time tlist))
      (setq loc_fts (format-time-string time_format enc_time (current-time-zone)))
      (setq thai_fts (format-time-string time_format enc_time thai_zone))
      (setq o_str (concat "thai time = " thai_fts "\n" "time      = " loc_fts)))
     ;; local time
     (t
      (setq tlist (append tlist (nthcdr 3 (decode-time c_time))))
      (setq enc_time (apply 'encode-time tlist))
      (setq loc_fts (format-time-string time_format enc_time (current-time-zone)))
      (setq thai_fts (format-time-string time_format enc_time thai_zone))
      (setq o_str (concat "time      = " loc_fts "\n" "thai time = " thai_fts))))
    (message o_str)))

(defun pthai-time-th2loc(timestr)
"convert a thai time (hour/minute/second) to local time.  use
\"now\" for current time"
  (interactive "stime: ")
  (pthai-time-loc2th timestr t))

(defun pthai-time-now(&optional loctime)
  "return thai current time.  if optional loctime, return current
local time. message user time if called interactively"
  (interactive "P")
  (let ((c_zone (current-time-zone))
	(t_zone (current-time-zone nil (* 7 3600)))
	(time_format (concat "%a %b %e %H:%M:%S %Z %Y"))
	tstr)
    (setq tstr
	  (format-time-string time_format (current-time)
			      (if loctime c_zone t_zone)))
    (if (called-interactively-p 'any)
	(message tstr))
    tstr))

(defun pthai-say-time(timestr)
  "say a time in thai, use time like 1pm, 12:45pm, 17:34.  use
\"now\" for current time. returns time string"
  (interactive "stime: ")
  (let (tlist hour minute mhour tmp)
    (while (not (and (setq tlist (pthai-parse-hour-minute-second timestr))
		     (setq minute (nth 1 tlist))
		     (setq hour (nth 2 tlist))))
      (setq timestr (read-string "time: ")))
    (setq mhour (mod hour 6))
    ;; get hour
    (cond
     ;; 12am = เที่ยงคืน or หกทุ่ม or สองยาม
     ((or (< hour 1) (and (= hour 24) (= minute 0)))
      (setq tmp (random 4))
      (cond ((< tmp 2)
	     (setq tlist (list "เที่ยงคืน")))
	    ((= tmp 2)
	     (setq tlist (list "หกทุ่ม")))
	    (t
	     (setq tlist (list "สองยาม")))))
     ;; 1am 2am 3am 4am 5am = ตี + number[1-5]
     ((< hour 6)
      (setq tlist (list "ตี" hour)))
     ;; 6am = หก โมง เช้า or ตีหก or ย่ำรุ่ง
     ((< hour 7)
      (setq tmp (random 4))
      (cond ((< tmp 2)
	     (setq tlist (list "หก" "โมง" "เช้า")))
	    ((= tmp 2)
	     (setq tlist (list "ตีหก")))
	    (t
	     (setq tlist (list "ย่ำรุ่ง")))))
     ;; 7am 8am 9am 10am 11am = 1-5  + โมง + เช้า
     ;;                       = 7-11 + โมง + (เช้า)
     ((< hour 12)
      (if (= 0 (random 2))     ;; 1-5
	  (if (= mhour 1)
	      (setq tlist (list "โมง" "เช้า"))
	    (setq tlist (list mhour "โมง" "เช้า")))
	(setq tlist (list hour "โมง"))  ;; 7-11
	(if (= 0 (random 2))
	    (setq tlist (append tlist (list "เช้า"))))))
     ;; 12pm = เที่ยงวัน
     ((< hour 13)
      (setq tlist (list "เที่ยงวัน")))
     ;; 1pm 2pm 3pm 4pm 5pm   = บ่าย + number[1-5] + โมง or
     ;;             4pm       = สี่ +  โมง + เย็น or
     ;;             5pm       = ห้า + โมง + เย็น
     ((< hour 18)
      (cond
       ((= hour 13) ;; 1pm
	(setq tlist (list "บ่าย" "โมง")))
       ((or (= hour 14) (= hour 15))
	(setq tlist (list "บ่าย" mhour "โมง")))
       ((= hour 16) ;; 4pm
	(if (= 0 (random 2))
	    (setq tlist (list "สี่" "โมง" "เย็น"))
	  (setq tlist (list "บ่าย" mhour "โมง"))))
       (t           ;; 5pm
	(if (= 0 (random 2))
	    (setq tlist (list "ห้า" "โมง" "เย็น"))
	  (setq tlist (list "บ่าย" mhour "โมง"))))))
     ;; 6pm = หก โมง เย็น or ย่ำค่ำ
     ((< hour 19)
      (if (= 0 (random 3))
	  (setq tlist (list "ย่ำค่ำ"))
	(setq tlist (list "หก" "โมง" "เย็น"))))
     ;; 7pm 8pm 9pm 10pm 11pm = number[1-5] + ทุ่ม
     ((< hour 24)
      (setq tlist (list mhour "ทุ่ม")))
     (t
      (error (concat "bad hour " hour))))
    ;; minutes are appended 10:07am = สิบ โมง (เช้า) + เจ็ด นาที
    (unless (or (= minute 0) (= hour 24))
      (setq tlist (append tlist (list minute "นาที"))))
    ;; convert everything to strings
    (setq tlist (mapcar (lambda (x) (if (numberp x) (number-to-string x) x))
			tlist))
    ;; play audio
    (dolist (tmp tlist)
      (pthai-say tmp))
    (setq tmp
	  (concat (number-to-string hour) ":" (format "%02d" minute)))
    (message tmp)
    tmp))

(defun pthai-parse-word-files(files)
  "take a list of files and directories, return full paths of
  files, and non-dotfiles in directories"
  (let (tmp dfiles rfiles)
    (while files
      (setq files (mapcar 'expand-file-name files))
      (setq files (sort (seq-uniq files) 'string-lessp))
      (setq tmp (car files))
      (setq files (cdr files))
      (when tmp
	(cond
	 ;; skip dot files
	 ((string-match "^\\." (file-name-nondirectory tmp))
	  nil)
	 ;; save list of files in dirs, skipping dot files and subdirs
	 ((file-directory-p tmp)
	  (setq dfiles (directory-files tmp nil "^\\([^.]\\|\\.[^.]\\|\\.\\..\\)"))
	  (setq dfiles (mapcar (lambda (x) (concat tmp "/" x)) dfiles))
	  (setq dfiles (seq-filter (lambda (x) (not (file-directory-p x))) dfiles))
	  (setq files (append files dfiles)))
	 ;; normal file
	 ((file-exists-p tmp)
	  (setq rfiles (append rfiles (list tmp)))))))
    (setq rfiles (sort (seq-uniq rfiles) 'string-lessp))))

(defun pthai-dictionary-add-word(thaiword &optional def english_class thai_class)
  "update pthai-dictionary with new entry or merge with existing
entry. thaiword is a thai string, other args can be nil, string,
or a list of strings. returns merged definition of word that was
inserted"
  (interactive)
  (let (ilist tmp)
    ;; set up lists if thaiword has any definition, else just add the
    ;; thaiword without any defs
    (if (or def english_class thai_class)
	(setq ilist (mapcar (lambda(x) (if (listp x) x (list x)))
			    (list def english_class thai_class)))
      (setq ilist '(nil nil nil)))
    ;; merge if entry already exists
    (if (setq tmp (gethash thaiword pthai-dictionary))
	(setq ilist (seq-mapn #'append ilist tmp)))
    ;; delete duplicates, nil
    (setq ilist
	  (mapcar (lambda(x) (sort (delq nil (delete-dups x)) 'string-lessp))
		  ilist))
    (puthash thaiword ilist pthai-dictionary)))

(defun pthai-dictionary-remove-word(thaiword)
  "remove word from dictionary, return t if word in dictionary,
  nil otherwise"
  (interactive "sthai word to remove: ")
  (cond ((gethash thaiword pthai-dictionary)
	 (remhash thaiword pthai-dictionary)
	 t)
	(t
	 nil)))

(defun pthai-parse-classifiers(str)
  "helper function to parse definition line and return list like
((a b c) (ก ด ห))"
  (let (eclass tclass tmp pos)
    (setq pos 0)
    (while (string-match "\\({.*?}\\)" str pos)
      (setq tmp (match-string 1 str))
      (setq pos (match-end 1))
      (if (string-match-p "\\ct+" tmp)
	  (setq tclass (append (split-string tmp " *[{} \t]+ *" t) tclass))
	(setq eclass (append (split-string tmp " *[{},\t]+ *" t) eclass))))
    (list eclass tclass)))

(defun pthai-dictionary-read(fname &optional clear_all read_twt)
  "read a dictionary file, adding words to pthai-dictionary hash.
if optional clear_dict arg, reset dictionary first.  if read_twt
arg, re-read the 'thai-word-table.  returns number of words read
in file.  assumes file is in utf-8 format and lines contain a
single thai word at beginning of the line, or a thai
\"definition\" line"
  (interactive "fpath to dictionary name: \n")
  (let ((coding-system-for-read 'utf-8)
	(coding-system-for-write 'utf-8)
	(buffer-file-coding-system 'utf-8)
	(count 0)
	tword def_line class_info)
    (unless (and (> (length fname) 0)
		 (setq fname (expand-file-name fname))
		 (file-exists-p fname)
		 (not (file-directory-p fname)))
      (error (concat "invalid dictionary \"" fname "\"")))
    (if clear_all
      (clrhash pthai-dictionary))
    (if read_twt
	(pthai-twt-read))
    (with-temp-buffer
      (insert-file-contents fname)
      ;; remove # comments
      (goto-char (point-min))
      (while (re-search-forward "#.*$" nil t)
	(replace-match ""))
      ;; remove stuff in parentheses
      (goto-char (point-min))
      (while (re-search-forward "[ \t]*(.*?)[ \t]*" nil t)
	(replace-match " "))
      ;; consolidate spaces
      (goto-char (point-min))
      (while (re-search-forward "[ \t]+" nil t)
	(replace-match " "))
      ;; remove double quote marks
      (goto-char (point-min))
      (while (re-search-forward "\"" nil t)
	(replace-match ""))
      ;; convert all to lowercase
      (downcase-region (point-min) (point-max))

      ;; read words without definitions
      (goto-char (point-min))
      (while (re-search-forward "^\\(\\ct+\\)$" nil t)
	(setq count (+ 1 count))
	(pthai-dictionary-add-word (match-string 1)))
      
      ;; read words with "dictionary" format
      (goto-char (point-min))
      (while (re-search-forward "^\\(\\ct+\\) \\(.*?\\) *$" nil t)
       	(when (and (setq tword (match-string 1))
       		   (setq def_line (match-string 2))
       		   (> (length def_line) 0))
	  (cond
	   ;; parse classifier info and strip from definition
	   ((string-match-p "[{}]" def_line)
	    (setq class_info (pthai-parse-classifiers def_line))
	    (setq def_line (replace-regexp-in-string " ?{.*?} ?" "" def_line)))
	   (t
	    (setq class_info nil)))
	  ;; save thai word, definition, classifier info	  
	  (pthai-dictionary-add-word tword
				     (split-string def_line " ?, ?" t)
				     (nth 0 class_info)
				     (nth 1 class_info))
	  (setq count (+ 1 count)))))
    (if pthai-verbose-wordloads
	(message "%d words in %s" count fname))
    count))

(defcustom pthai-dictionary-list
  (list (concat pthai-default-directory "dictionary"))
  "list of files and directories that contain thai dictionary files"
  :group 'pthai
  :set (lambda (sym val)
	 (set-default sym val)
	 (let (tmp)
	   (dolist (tmp (pthai-parse-word-files val))
	     (if (file-exists-p tmp)
		 (pthai-dictionary-read tmp)))
	   (unless pthai-bootstrap
	     (when pthai-verbose-wordloads
	       (setq tmp (hash-table-count pthai-dictionary))
	       (message "%d unique words in pthai-dictionary" tmp)))))
  :type '(repeat file))

(defun pthai-dictionary-readall(&optional clear)
  "read dictionary files defined in pthai-dictionary-list.  if
optional clear, clear existing entries.  returns total number of
words in dictionary"
  (interactive "P")
  (let (tmp)
    (when clear
      (clrhash pthai-dictionary)
      (pthai-twt-read))      
    (dolist (tmp (pthai-parse-word-files pthai-dictionary-list))
      (pthai-dictionary-read tmp))
    (when pthai-verbose-wordloads
      (setq tmp (hash-table-count pthai-dictionary))
      (message "%d unique words in pthai-dictionary" tmp))
    tmp))

(defun pthai-word-count()
  "return number of words in dictionary.  message user when
called interactively, displaying counts of words with and without
definitions"
  (interactive)
  (let ((total (length (hash-table-keys pthai-dictionary)))
	(defined 0)
	(undefined 0))
    (dolist (tmp (hash-table-keys pthai-dictionary))
      (if (pthai-dictionary-undef-p tmp)
	  (setq undefined (+ 1 undefined))
	(setq defined (+ 1 defined))))
    (when (called-interactively-p 'any)
      (message
       "defined/undefined/total = %d/%d/%d"
       defined undefined total))
    total))

(defun pthai-reverse-lookup(word ltype)
  "helper function for reverse look ups:
  word                 ltype      returns (list of ... or nil if not found)       example
  ====                 =====      ================                                =======
  english word         'def       thai definitions                                dog => (สุนัข หมา)
  english classifier   'eclass    thai words that are classified by english word  laws => (กฎ)
  thai classifier      'tclass    thai words that are classified by thai word     วงกบ => (กรอบ วง)"
  (let* (tmp vals rlist
	     (ltypes (list 'def 'eclass 'tclass))
	     (n (seq-position ltypes ltype)))
    (unless n
      (error "invalid ltype: %s" ltype))
    (dolist (tmp (hash-table-keys pthai-dictionary))
	(setq vals (nth n (gethash tmp pthai-dictionary)))
	(if (member word vals)
	    (setq rlist (append rlist (list tmp)))))
    (setq rlist (sort (delete-dups rlist) 'string-lessp))))

(defun pthai-dictionary-p(thaiword)
  "return non-nil if thaiword in dictionary, regardless of
whether it has a definition or not"
  (if (gethash thaiword pthai-dictionary)
      t nil))

(defun pthai-dictionary-undef-p(thaiword)
  "return non-nil if thaiword in dictionary and is without any
definition. returns nil if not in dictionary or in dictionary and
has any definition"
  (let (tmp)
    (if (and (setq tmp (gethash thaiword pthai-dictionary))
	     (not (nth 0 tmp))
	     (not (nth 1 tmp))
	     (not (nth 2 tmp)))
	t nil)))

(defun pthai-classifier-p(thaiword)
  "return non-nil if thaiword is a classifier in dictionary"
  (let (tmp)
    (if (and (setq tmp (gethash thaiword pthai-dictionary))
	     (or (nth 1 tmp) (nth 2 tmp)))
	t nil)))

(defun pthai-find-word-regexp(thairegexp)
  "return list of thai words from words in dictionary matching a
regexp using string-match"
  (interactive "sthai regexp: \n")
  (let (tmp words)
    (with-temp-buffer
      (dolist (tmp (hash-table-keys pthai-dictionary))
	(if (string-match-p thairegexp tmp)
	    (insert tmp "\n")))
      (sort-lines nil (point-min) (point-max))
      (setq words (split-string (buffer-string) "\n" t)))
      
    (when (called-interactively-p 'any)
      (if words
	  (message "%s" (string-join words " "))
	(message "no matches found for %s" thairegexp)))
    words))

(defun pthai-lookup(word &optional force_message)
  "look up definition of thai or english word. returns list of
english or thai words.  if called interactively or forced,
message user definition"
  (interactive "sword: \n")
  (let (wlist)
    (setq word (downcase word))
    (if (string-match-p "\\ct+" word)
	(setq wlist (nth 0 (gethash word pthai-dictionary)))
      (setq wlist (pthai-reverse-lookup word 'def)))
    (if (and (not wlist)
	     (string-match-p "^[0-9๐๑๒๓๔๕๖๗๘๙]+$" word))
	(setq wlist (list (pthai-number word))))
    (when (or (called-interactively-p 'any) force_message)
      (if wlist
	  (message "%s = %s" word (string-join wlist ", "))
	(message "%s = ?" word)))
    wlist))

(defun pthai-lookup-at-point()
  "same as pthai-lookup but do word at point"
  (interactive)
  (let ((bounds (or (pthai-bounds-of-thaiword-at-point)
		    (bounds-of-thing-at-point 'word))))
    (unless bounds
      (error "could not determine bounds"))
    (pthai-highlight-say-bounds bounds nil t)))

(defun pthai-lookup-and-insert(word)
  "look up a word (english or thai) and insert meaning at point.
if more than one definition found, select from choices found"
  (interactive "sword: ")
  (let (tlist tlen)
    (save-excursion
      (setq word (downcase word)
	    tlist (pthai-lookup word)
	    tlen (length tlist))
      (cond
       ((= 0 tlen)
	(message "%s not found in dictionary" word))
       ((= 1 tlen)
	(insert (car tlist)))
       (t
	(insert (ido-completing-read "meanings: " tlist)))))))

(defun pthai-lookup-classifier-detail(word)
  "return detailed classifier information for word. see
'pthai-lookup-classifier"
  (interactive "sword to lookup: \n")
  (pthai-lookup-classifier word t))

(defun pthai-lookup-classifier(word &optional detail)
    "return classifier information for word:
word                      return value (list of ...)
====                      ============
thai word, classifier     '( (defintions) '(eng classifiers) '(thai classifiers) ) 
thai word, not classifer  '( thai classifiers for thai word )
english word              '( thai classifiers for english word )

    message user information if called interactively. if optional
detail arg given, print out complete definitions for all
classifier words (for 2nd and 3rd cases, 1st case will by default
when interactive"
  (interactive "sword to lookup: \n")
  (let (tmp clist ret)
    (setq word (downcase word))
    (cond
     ;; thai word that is a classifier, e.g. ตัว
     ((pthai-classifier-p word)
      (setq ret (gethash word pthai-dictionary))
      (setq clist (list word)))
     ;; thai word that has classifiers, e.g. แมว
     ((string-match-p "\\ct+" word)
      (setq ret (pthai-reverse-lookup word 'tclass))
      (setq clist ret))
     ;; english word, e.g. dog
     ((string-match-p "^\\Ct+$" word)
      ;; look up classifiers directly and reverse lookup thai defs and
      ;; the classifiers for them
      (setq clist (pthai-reverse-lookup word 'eclass))
      (dolist (tmp (pthai-lookup word))
	(setq clist (append (pthai-reverse-lookup tmp 'tclass) clist)))
      (setq clist (seq-filter (lambda (x) x) clist))
      (setq clist (sort (delete-dups clist) 'string-lessp))
      (setq ret clist))
     (t
      (error "error parsing classifier info")))

    (when (or detail (called-interactively-p 'any))
      (cond
       ;; word is a classifier, print out info
       ((pthai-classifier-p word)
	(message (pthai-classifier-info word)))
       ;; found classifiers for word, print list of classifiers, or details
       ;; for each classifier found
       (clist
	(if detail
	    (message (mapconcat 'pthai-classifier-info clist "\n"))
	  (message "%s classified by: %s" word (string-join clist ", "))))
       ;; no classifier info found
       (t
	(message (pthai-classifier-info word)))))
    ret))

(defun pthai-classifier-info(word)
  "return classifer information as a string"
  (let (cinfo cistr)
    (if (not (pthai-classifier-p word))
	(setq cistr (concat "no classifier info found for " word))

      (setq cinfo (gethash word pthai-dictionary))      
      (if (nth 0 cinfo)
	  (setq cistr (concat word " = " (string-join (nth 0 cinfo) ", ") "\n"))
	(setq cistr (concat word " is a classifier\n")))
      (if (nth 1 cinfo)
	  (setq cistr (concat cistr "classifier for: "
			     (string-join (nth 1 cinfo) ", ") "\n")))
      (if (nth 2 cinfo)
	  (setq cistr (concat cistr "classifier for: "
			     (string-join (nth 2 cinfo) " ") "\n")))
      (setq cistr (string-trim-right cistr)))
    cistr))

(defun pthai-temperature-cel2fah(celsius)
  "convert celsius to fahrenheit.  message user if called
interactively"
  (interactive "ncelsius: ")
  (let ((fah (+ (* celsius (/ 9.0 5.0)) 32.0)))
    (if (called-interactively-p 'any)
	(message "%3.1f C = %3.1f F" celsius fah))
    fah))

(defun pthai-temperature-fah2cel(fahrenheit)
  "convert fahrenheit to celsius. message user if called
interactively"
  (interactive "nfahrenheit: ")
  (let ((cel (* (- fahrenheit 32.0) (/ 5.0 9.0))))
    (if (called-interactively-p 'any)
	(message "%3.1f F = %3.1f C" fahrenheit cel))
    cel))

(defun pthai-year-greg2bud(&optional year)
  "convert gregorian year (A.D.) to buddhist year (B.E.). message
user if called interactively"
  (interactive)
  (let (bud_year
	(cur_year (string-to-number (format-time-string "%Y"))))
    (unless (integerp year)
      (setq year (round (read-number "year: " cur_year))))
    (setq bud_year (+ 543 year))
    (when (called-interactively-p 'any)
      (if (>= year 0)
	  (message "%d A.D. = B.E. %d" year bud_year)
	(message "%d B.C. = B.E. %d" (* -1 year) bud_year)))
    bud_year))

(defun pthai-year-bud2greg(&optional year)
  "convert buddhist year (B.E.) to gregorian year (A.D./B.C.).
message user if called interactively"
  (interactive)
  (let (greg_year
	(cur_year (+ 543
		     (string-to-number (format-time-string "%Y")))))
    (unless (integerp year)
      (setq year (round (read-number "year: " cur_year))))
    (setq greg_year (- year 543))
    (when (called-interactively-p 'any)
      (if (>= greg_year 0)
	  (message "B.E. %d = %d A.D." year greg_year)
	(message "B.E. %d = %d B.C. " year (* -1 greg_year))))
    greg_year))

(defun pthai-x-rates-currencies()
 "from www.x-rates.com, get an alist of supported currencies and their names
'((\"USD\" . \"US Dollar\") (\"THB\" . \"Thai Baht\") ... )"
  (let ((x_url "https://www.x-rates.com")
	clist
	;; match a string like
	;;<li><a href='https://www.x-rates.com/table/?from=ARS' onclick="submitConverterArgs(this)" rel='ratestable'>Argentine Peso</a></li>
	(currency_rexp
	 (concat "<li><a href='https://www.x-rates.com/table/\\?from=\\([A-Z]\\{3\\}\\)' "
		 "onclick=.* rel='ratestable'>\\(.*\\)</a></li>")))
    (with-current-buffer (url-retrieve-synchronously x_url)
      (goto-char (point-min))
      (while (re-search-forward currency_rexp nil t)
	(setq clist (cons (cons (match-string 1) (match-string 2)) clist))))
    (unless (> (length clist) 40)
      (error "error fetching currency types"))
    (reverse clist)))

(defun pthai-x-rates-convert(from to amount)
  "convert a currency (USD THB ...) amount to another currency,
from x-rates.com.  returns converted amount as a number.  message
user conversion info if called interactively"
  (let ((x_url "https://www.x-rates.com")
	(curr_alist (pthai-x-rates-currencies))
	from_name
	to_name
	url
	mstr
	converted)
    (unless (and amount (numberp amount) (> amount 0.0))
      (error "invalid monetary amount"))
    (setq amount (number-to-string amount))
    ;; sanity check dollar & baht, and then look up names
    (unless (and (assoc "USD" curr_alist)
		 (assoc "THB" curr_alist)
		 (setq to_name (cdr (assoc to curr_alist)))
		 (setq from_name (cdr (assoc from curr_alist))))
      (error (concat "unknown currency to/from type(s): to=" to " from=" from)))
    ;; fetch current rates, then match something like
    ;; <a href='https://www.x-rates.com/graph/?from=USD&amp;to=THB'>30.171404</a>
    (setq url (concat x_url "/table/?from=" from "&amount=" amount))
    (setq mstr (concat "?from=" from "&amp;to=" to "'>"))
    (with-current-buffer (url-retrieve-synchronously url)
      (goto-char (point-min))
      (and (search-forward mstr nil t)
	   (re-search-forward "\\([0-9\.]+\\)" nil t)
	   (setq converted (format "%3.2f" (string-to-number (match-string 1))))))
    (unless converted
      (error (concat "download conversion match not found for " url)))
    (if (not (or executing-kbd-macro noninteractive))
	(message "%s %s = %s %s" amount from_name converted to_name))
    (string-to-number converted)))

(defun pthai-currency-dollar-to-baht(&optional amount)
  "convert us dollars to baht"
  (interactive)
  (unless (numberp amount)
    (setq amount (read-number "dollars: " 1)))
  (pthai-x-rates-convert "USD" "THB" amount))

(defun pthai-currency-baht-to-dollar(&optional amount)
  "convert baht to us dollars"
  (interactive)
  (unless (numberp amount)
    (setq amount (read-number "baht: " 1)))
  (pthai-x-rates-convert "THB" "USD" amount))

(defun pthai-currency-dollar-to-euro(&optional amount)
  "convert us dollars to euros"
  (interactive)
  (unless (numberp amount)
    (setq amount (read-number "dollars: " 1)))
  (pthai-x-rates-convert "USD" "EUR" amount))

(defun pthai-currency-euro-to-dollar(&optional amount)
  "convert euros to us dollars"
  (interactive)
  (unless (numberp amount)
    (setq amount (read-number "euros: " 1)))
  (pthai-x-rates-convert "EUR" "USD" amount))

(defun pthai-currency-dollar-to-pound(&optional amount)
  "convert us dollars to british pounds"
  (interactive)
  (unless (numberp amount)
    (setq amount (read-number "dollars: " 1)))
  (pthai-x-rates-convert "USD" "GBP" amount))

(defun pthai-currency-pound-to-dollar(&optional amount)
  "convert british pounds to us dollars"
  (interactive)
  (unless (numberp amount)
    (setq amount (read-number "pounds: " 1)))
  (pthai-x-rates-convert "GBP" "USD" amount))

;; initialize default dictionary
(unless pthai-bootstrap
  (pthai-twt-update "ST_SHARE_DIR/words" t)
  ;; to prevent "Variable binding depth exceeds max-specpdl-size"
  (setq max-specpdl-size 2500)
  (pthai-twt-read t)
  (pthai-dictionary-read "ST_SHARE_DIR/sampledict.txt"))

(provide 'pthai)
